﻿--------------------------------------------------------------------------------
-- Module Declaration
--

local mod = BigWigs:NewBoss("Garajal Cross Over", 896)
if not mod then return end
mod:RegisterEnableMob(60143)
-- XXX make this local once done debugging
LGIST = LibStub:GetLibrary("LibGroupInSpecT-1.0")

--------------------------------------------------------------------------------
-- Locals
--
local DPStoCrossOver = 3 -- changed in :OnEngage
local HealerToCrossOver = 1
local totemCounter = 1
local frailSoul, voodooDoll, crossedOver, conduitOn, spiritualInnervation = (GetSpellInfo(117723)), (GetSpellInfo(122151)), (GetSpellInfo(116161)), (GetSpellInfo(122181)), (GetSpellInfo(117549))
local potentialTargets = {}
local lastCrossedOver = {}
local mainMod = BigWigs.modules.Bosses.modules["Gara'jal the Spiritbinder"]
local crossOverTargets = mod:NewTargetList()
local diff, numRaid


--------------------------------------------------------------------------------
-- Localization
--

local ML = mainMod:GetLocale("enUS", true)

local L = mod:NewLocale("enUS", true)
if L then
	L.cross_over = "Cross Over"
	L.cross_over_desc = "Warning for who should cross over with the most recent totem."
	L.cross_over_icon = 116161
end
L = mod:GetLocale()

--------------------------------------------------------------------------------
-- Initialization
--

function mod:GetOptions()
	return {
		{"cross_over", "FLASHSHAKE"}, "bosskill",
	}, {
		cross_over = "general",
	}
end

function mod:OnBossEnable()
	self:RegisterEvent("INSTANCE_ENCOUNTER_ENGAGE_UNIT", "CheckBossStatus")

	self:Death("Win", 60143)
end

function mod:OnEngage()
	totemCounter = 1
	wipe(potentialTargets)
	wipe(lastCrossedOver)
	local numRaid = GetNumGroupMembers()
	local diff = GetRaidDifficulty()
	if diff == 2 or diff == 4 then
		DPStoCrossOver = 3
	else
		DPStoCrossOver = 2
	end
	for i=1, numRaid do
		local name, _, subgroup = GetRaidRosterInfo(i)
		if diff == 2 or diff == 4 then
			if subgroup < 6 then
				potentialTargets[name] = true
			end
		else
			if subgroup < 3 then
				potentialTargets[name] = true
			end
		end
	end
end

--------------------------------------------------------------------------------
-- Event Handlers
--

local function isDebuffed(player)
	if UnitExists(player) then
		if UnitDebuff(player, frailSoul) or UnitDebuff(player, voodooDoll) or UnitDebuff(player, crossedOver) or UnitDebuff(player, conduitOn) then
			return true
		else
			return nil
		end
	end
end

do
	local f = CreateFrame("Frame")
	local total = 0
	f:SetScript("OnUpdate", function(self, elapsed)
		total = total + elapsed
		if total > 0.05 then
			local numRaid = GetNumGroupMembers()
			local diff = GetRaidDifficulty()
			-- keep track of debuffs
			for i=1, numRaid do
				local name, _, subgroup = GetRaidRosterInfo(i)
				-- keep track of when was crossed over last seen on the target
				if UnitExists(name) then
					if UnitDebuff(name, crossedOver) then
						lastCrossedOver[name] = GetTime() -- because debuff can be removed earlier then it's duration
					end
				end
				-- check if there are any active debuffs on eligible targets
				if diff == 2 or diff == 4 then
					if subgroup < 6 then
						if isDebuffed(name) then
							potentialTargets[name] = false
						else
							potentialTargets[name] = true
						end
					end
				else
					if subgroup < 3 then
						if isDebuffed(name) then
							potentialTargets[name] = false
						else
							potentialTargets[name] = true
						end
					end
				end
			end

			-- now check if we are in heroic, if so, handle frail soul being applied late
			if mod:Heroic() then
				for k, v in pairs(lastCrossedOver) do
					if (GetTime() - v) < 1 then -- less than 1 sec ago still had corssed over
						potentialTargets[k] = v + 3 -- should have frail soul by now
					end
				end
			end
		end
	end)
end

function mainMod:OnSync(sync, rest)
	if sync == "Totem" and rest then
		self:Message(116174, ("%s (%d)"):format(ML["totem"], totemCounter), "Attention", 116174)
		totemCounter = totemCounter + 1
		if self:Heroic() then
			self:Bar(116174, ("%s (%d)"):format(ML["totem"], totemCounter), 20, 116174)
		else
			self:Bar(116174, ("%s (%d)"):format(ML["totem"], totemCounter), 36, 116174)
		end
		-- sort the table
		local sortedTable = {}
		local lowManaHealersAssigned = {}
		for k, v in pairs(potentialTargets) do
			if v then
				-- check for those recently crossed over back to the living world
				if type(v) == "number" then -- this should almost assure we are on heroic
					if mod:Heroic() then -- but this makes sure when we go from heroic to normal nothing unexpected happens
						if (GetTime() - potentialTargets[k]) > 0 then -- the 3 sec have passed since crossed over faded
							-- recheck if the person has debuffs
							if not isDebuffed(k) then
								table.insert(sortedTable, k)
							end
						end
					end
				else -- boolean
					table.insert(sortedTable, k)
				end
			end
		end
		table.sort(sortedTable, function(a,b) return a<b end)
		local healerCounter, DPScounter = 0, 0
		for k, player in ipairs(sortedTable) do
			-- assign non disc priest low mana healers
			if not UnitIsDeadOrGhost(player) and UnitIsConnected(player) then
				local info = LGIST:GetCachedInfo(UnitGUID(player))
				if healerCounter < HealerToCrossOver then
					if info.spec_role and info.spec_role == "HEALER" then
						if not (info.class == "PRIEST" and info.spec_index == 1) then -- not disc priest
							local maxMana = UnitPowerMax(player)
							local currMana = UnitPower(player)
							if maxMana and currMana then
								if (currMana/maxMana*100) < 25 then -- low mana
									lowManaHealersAssigned[player] = true
									crossOverTargets[#crossOverTargets+1] = player
									healerCounter = healerCounter + 1
									if UnitIsUnit("player", player) then
										mod:FlashShake("cross_over")
									end
								end
							end
						end
					end
				end
				-- first try and assign ranged dps without the dps buff
				if DPScounter < DPStoCrossOver then
					if not UnitBuff(player, spiritualInnervation) then
						if info.spec_role and info.spec_role == "DAMAGER" then
							if info.class == "HUNTER" or info.class == "MAGE" or info.class == "PRIEST" or info.class == "WARLOCK" or
						(info.class == "DRUID" and info.spec_index == 1) or (info.class == "SHAMAN" and info.spec_index == 1) then
								crossOverTargets[#crossOverTargets+1] = player
								DPScounter = DPScounter + 1
								if UnitIsUnit("player", player) then
									mod:FlashShake("cross_over")
								end
							end
						end
					end
				end
			end
		end
		-- get the higher mana healers
		for k, player in ipairs(sortedTable) do
			if not UnitIsDeadOrGhost(player) and UnitIsConnected(player) then
				local info = LGIST:GetCachedInfo(UnitGUID(player))
				if healerCounter < HealerToCrossOver then
					if info.spec_role and info.spec_role == "HEALER" then
						if not (info.class == "PRIEST" and info.spec_index == 1) then -- non disc
							if not lowManaHealersAssigned[player] then -- not an already assigned healer
								crossOverTargets[#crossOverTargets+1] = player
								healerCounter = healerCounter + 1
								if UnitIsUnit("player", player) then
									mod:FlashShake("cross_over")
								end
							end
						end
					end
				end
			end
		end
		-- get disc priests now since nothing else left
		for k, player in ipairs(sortedTable) do
			if not UnitIsDeadOrGhost(player) and UnitIsConnected(player) then
				local info = LGIST:GetCachedInfo(UnitGUID(player))
				if healerCounter < HealerToCrossOver then
					if info.spec_role and info.spec_role == "HEALER" then
						if info.class == "PRIEST" and info.spec_index == 1 then -- disc
							crossOverTargets[#crossOverTargets+1] = player
							healerCounter = healerCounter + 1
							if UnitIsUnit("player", player) then
								mod:FlashShake("cross_over")
							end
						end
					end
				end
			end
		end
		-- assign ranged with the dps buff
		for k, player in ipairs(sortedTable) do
			if not UnitIsDeadOrGhost(player) and UnitIsConnected(player) and UnitBuff(player, spiritualInnervation) then
				local info = LGIST:GetCachedInfo(UnitGUID(player))
				if DPScounter < DPStoCrossOver then
					if info.spec_role and info.spec_role == "DAMAGER" then
						if info.class == "HUNTER" or info.class == "MAGE" or info.class == "PRIEST" or info.class == "WARLOCK" or
						(info.class == "DRUID" and info.spec_index == 1) or (info.class == "SHAMAN" and info.spec_index == 1) then
							crossOverTargets[#crossOverTargets+1] = player
							DPScounter = DPScounter + 1
							if UnitIsUnit("player", player) then
								mod:FlashShake("cross_over")
							end
						end
					end
				end
			end
		end
		-- fill in remaining dps slots with melee
		for k, player in ipairs(sortedTable) do
			if not UnitIsDeadOrGhost(player) and UnitIsConnected(player) then
				local info = LGIST:GetCachedInfo(UnitGUID(player))
				if DPScounter < DPStoCrossOver then
					if info.spec_role and info.spec_role == "DAMAGER" then
						if info.class == "DEATHKNIGHT" or info.class == "MONK" or info.class == "WARRIOR" or info.class == "ROGUE" or
						info.class == "PALADIN" or (info.class == "DRUID" and info.spec_index == 2) or (info.class == "SHAMAN" and info.spec_index == 2) then
							crossOverTargets[#crossOverTargets+1] = player
							DPScounter = DPScounter + 1
							if UnitIsUnit("player", player) then
								mod:FlashShake("cross_over")
							end
						end
					end
				end
			end
		end
		-- one last pass to fill up the list based on blizz role assignments, in case the lib talent scanning supplied no data
		for k, player in ipairs(sortedTable) do
			if not UnitIsDeadOrGhost(player) and UnitIsConnected(player) then
				local info = LGIST:GetCachedInfo(UnitGUID(player))
				if healerCounter < HealerToCrossOver then
					if not info.spec_role and UnitGroupRolesAssigned(player) == "HEALER" then
						crossOverTargets[#crossOverTargets+1] = player
						healerCounter = healerCounter + 1
						if UnitIsUnit("player", player) then
							mod:FlashShake("cross_over")
						end
					end
				end
				if DPScounter < DPStoCrossOver then
					if info.spec_role and UnitGroupRolesAssigned(player) == "DAMAGER" then
						crossOverTargets[#crossOverTargets+1] = player
						DPScounter = DPScounter + 1
						if UnitIsUnit("player", player) then
							mod:FlashShake("cross_over")
						end
					end
				end
			end
		end

		if healerCounter < HealerToCrossOver then
			print(("Healer assigned to cross over: %d (which might be low, tried to assign %d)"):format(healerCounter, HealerToCrossOver))
		end
		if DPScounter < DPStoCrossOver then
			print(("DPS assigned to cross over: %d (which might be low, tried to assign %d)"):format(DPScounter, DPStoCrossOver))
		end
		mod:TargetMessage("cross_over", L["cross_over"], crossOverTargets, "Important", 116161, "Info")
	end
end

